"""The load unload problem. An agent is placed on a one dimensional grid world
and is tasked with loading itself up on the right side of the world and
unloading on the left. The agent can observe whether or not it is in the load or
unload block but can not tell its exact location of whether it is loaded or
unloaded. Therefore the agent must maintain belief about it's location and load
status.

States are defined by the location of the agent and whether or not it is loaded
Actions: "move-left", "move-right"
Rewards:
    +100 for moving into the unload block while loaded
    -1 otherwise

"""

import pomdp_py
import random
import numpy as np
import sys
import copy
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation


EPSILON = 1e-3
LOAD_LOCATION = 10

class LUState(pomdp_py.State):
    def __init__(self, x, loaded):
        if type(x) != int or x < 0:
            raise ValueError("Invalid state: {}\n".format((x, loaded)) +
                             "x must be an integer > 0")
        if type(loaded) != bool:
            raise ValueError("Invalid state: {}\n".format((x, loaded)) +
                             "loaded must be a boolean")
        if x == 0 and loaded == True:
            raise ValueError("Agent can not be loaded in the 0th position")
        if x == LOAD_LOCATION and loaded == False:
            raise ValueError("Agent can not be unloaded in the last position")

        self.x = x
        self.loaded = loaded
    def __hash__(self):
        return hash((self.x, self.loaded))
    def __eq__(self, other):
        if isinstance(other, LUState):
            return self.x == other.x and self.loaded == self.loaded
        elif type(other) == tuple:
            return self.x == other[0] and self.loaded == other[1]
    def __str__(self):
        return str((self.x, self.loaded))
    def __repr__(self):
        return "State({})".format(self)

class LUAction(pomdp_py.Action):
    def __init__(self, name):
        if name not in ["move-left", "move-right"]:
            raise ValueError("Invalid action: %s" % name)

        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, LUAction):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return self.name
    def __repr__(self):
        return "Action(%s)" % self.name

class LUObservation(pomdp_py.Observation):
    def __init__(self, obs):
        if obs not in ["load", "unload", "middle"]:
            raise ValueError("Invalid observation: {}\n".format(name) +
                             "Observation must be an integer > 0")
        self.name = obs
    def __hash__(self):
        return hash(self.name)
    def __eq__(self, other):
        if isinstance(other, LUObservation):
            return self.name == other.name
        elif type(other) == str:
            return self.name == other
    def __str__(self):
        return str(self.name)
    def __repr__(self):
        return "Observation(%s)" % str(self.x)

# Observation model
class LUObservationModel(pomdp_py.ObservationModel):
    """This problem is small enough for the probabilities to be directly given
    externally"""
    def probability(self, observation, next_state, action, normalized=False, **kwargs):
        if observation != self.sample(next_state, action):
            # return EPSILON to avoid degradation of particles
            return EPSILON
        else:
            return 1 - EPSILON

    def sample(self, next_state, action, normalized=False, **kwargs):
        if next_state.x == 0:
            return LUObservation("unload")
        elif next_state.x == LOAD_LOCATION:
            return LUObservation("load")
        else:
            return LUObservation("middle")

    def argmax(self, next_state, action, normalized=False, **kwargs):
        """Returns the most likely observation"""
        return self.sample(next_state, action)


# Transition Model
class LUTransitionModel(pomdp_py.TransitionModel):
    """This problem is small enough for the probabilities to be directly given
            externally"""
    def probability(self, next_state, state, action, normalized=False, **kwargs):
        if next_state != self.sample(state, action):
            return EPSILON
        else:
            return 1 - EPSILON

    def sample(self, state, action, normalized=False, **kwargs):
        if ((state.x == LOAD_LOCATION and action == "move-right") or
           (state.x == 0 and action == "move-left")):
            # trying to make invalid move, stay in the same place
            return state

        if action == "move-right":
            # make sure we're always loaded in the far right cell
            if state.x == LOAD_LOCATION - 1:
                return LUState(state.x + 1, True)
            return LUState(state.x + 1, state.loaded)

        if action == "move-left":
            # make sure we're always unloaded in the first cell
            if state.x == 1:
                return LUState(state.x - 1, False)
            return LUState(state.x - 1, state.loaded)

    def argmax(self, state, action, normalized=False, **kwargs):
        """Returns the most likely next state"""
        return self.sample(state, action)

# Reward Model
class LURewardModel(pomdp_py.RewardModel):
    def probability(self, reward, state, action, next_state, normalized=False, **kwargs):
        if reward == self.sample(state, action):
            return 1.0
        else:
            return 0.0

    def sample(self, state, action, next_state, normalized=False, **kwargs):
        # if we are unloaded things, give reward 100, otherwise give -1
        if action == "move-left" and state.loaded == True and state.x == 1:
            return 100
        else:
            return -1

    def argmax(self, state, action, next_state, normalized=False, **kwargs):
        """Returns the most likely reward"""
        return self.sample(state, action)

# Policy Model
class LUPolicyModel(pomdp_py.RandomRollout):
    """This is an extremely dumb policy model; To keep consistent
    with the framework."""
    def __init__(self):
        self._all_actions = {LUAction("move-right"), LUAction("move-left")}

    def probability(self, action, state, normalized=False, **kwargs):
        raise NotImplementedError  # Never used

    def sample(self, state, normalized=False, **kwargs):
        return self.get_all_actions().random()

    def argmax(self, state, normalized=False, **kwargs):
        """Returns the most likely reward"""
        raise NotImplementedError

    def get_all_actions(self, **kwargs):
        return self._all_actions


class LoadUnloadProblem(pomdp_py.POMDP):

    def __init__(self, init_state, init_belief):
        """init_belief is a Distribution."""
        agent = pomdp_py.Agent(init_belief,
                               LUPolicyModel(),
                               LUTransitionModel(),
                               LUObservationModel(),
                               LURewardModel())

        env = pomdp_py.Environment(init_state,
                                   LUTransitionModel(),
                                   LURewardModel())

        super().__init__(agent, env, name="LoadUnloadProblem")

def generate_random_state():
    # Flip a coin to determine if we are loaded
    loaded = np.random.rand() > 0.5
    location = np.random.randint(0, LOAD_LOCATION + 1)
    if location == 0:
        loaded = False
    if location == LOAD_LOCATION:
        loaded = True
    return LUState(location, loaded)

def generate_init_belief(num_particles):
    particles = []
    for _ in range(num_particles):
        particles.append(generate_random_state())

    return pomdp_py.Particles(particles)

def test_planner(load_unload_problem, planner, nsteps=3, discount=0.95):
    gamma = 1.0
    total_reward = 0
    total_discounted_reward = 0

    fig = plt.figure()
    plt.title("Load/Unload problem (Red = empty, Blue = full)")
    plt.xlabel("Position")

    ax = fig.add_subplot(111)
    ax.set_xlim(-1, LOAD_LOCATION+1)
    ax.set_ylim(0, 2)
    x, y = [], []
    scat, = ax.plot(x, y, marker="x", markersize=20, ls=" ", color="black")

    def update(t):
        nonlocal gamma, total_reward, total_discounted_reward
        print("==== Step %d ====" % (t+1))
        action = planner.plan(load_unload_problem.agent)

        env_reward = load_unload_problem.env.state_transition(action, execute=True)
        true_state = copy.deepcopy(load_unload_problem.env.state)

        real_observation = load_unload_problem.env.provide_observation(
                load_unload_problem.agent.observation_model, action)
        load_unload_problem.agent.update_history(action, real_observation)
        planner.update(load_unload_problem.agent, action, real_observation)
        total_reward += env_reward
        total_discounted_reward += env_reward * gamma
        gamma *= discount
        print("True state: %s" % true_state)
        print("Action: %s" % str(action))
        print("Observation: %s" % str(real_observation))
        print("Reward: %s" % str(env_reward))
        print("Reward (Cumulative): %s" % str(total_reward))
        print("Reward (Cumulative Discounted): %s" % str(total_discounted_reward))
        print("Belief: %s" % str(load_unload_problem.agent.sample_belief()))
        if isinstance(planner, pomdp_py.POUCT):
            print("__num_sims__: %d" % planner.last_num_sims)
            print("__plan_time__: %.5f" % planner.last_planning_time)
        if isinstance(planner, pomdp_py.PORollout):
            print("__best_reward__: %d" % planner.last_best_reward)

        new_x, new_y = [true_state.x], [1]
        scat.set_data(new_x, new_y)
        scat.set_color("b" if true_state.loaded else "r")
        return scat,

    ani = FuncAnimation(fig, update, frames=nsteps, interval=500)
    plt.show()

def main():
    init_state = generate_random_state()
    init_belief = generate_init_belief(num_particles=100)
    load_unload_problem = LoadUnloadProblem(init_state, init_belief)

    print("** Testing POMCP **")
    pomcp = pomdp_py.POMCP(max_depth=100, discount_factor=0.95,
                           num_sims=100, exploration_const=110,
                           rollout_policy=load_unload_problem.agent.policy_model)
    test_planner(load_unload_problem, pomcp, nsteps=100)

if __name__ == '__main__':
    main()



# Assorted convenient functions
from pomdp_py.utils.math import (vec, proj,
                                 R_x, R_y, R_z, R_between, T,
                                 to_radians, approx_equal)
from pomdp_py.utils.misc import (remap, json_safe, safe_slice, similar, special_char)
from pomdp_py.utils.plotting import (plot_points, save_plot)
from pomdp_py.utils.colors import (lighter, rgb_to_hex, hex_to_rgb,
                                   inverse_color_rgb, inverse_color_hex,
                                   random_unique_color)
from pomdp_py.utils import typ
from pomdp_py.utils.debugging import TreeDebugger



import pomdp_problems.light_dark as ld
import matplotlib.pyplot as plt
import pomdp_py
import numpy as np
import time

count = 0

# true location.
x_0 = np.array([2.5, 0.0])
goal_pos = np.array([0.0, 0.0])

# defines the observation noise equation.
light = 5
const = 1

# planning horizon
planning_horizon = 30

# number of segments for direct transcription
num_segments = 10

var_sysd = 1e-9

# Environment.
env = ld.LightDarkEnvironment(x_0,  # init state
                              light,  # light
                              const)  # const
obsmodel = ld.ObservationModel(light, const)

func_sysd = env.transition_model.func()
func_obs = obsmodel.func()
jac_sysd = env.transition_model.jac_dx()
jac_sysd_u = env.transition_model.jac_du()
jac_obs = obsmodel.jac_dx()
noise_obs = obsmodel.func_noise()
noise_sysd = env.transition_model.func_noise(var_sysd)

L = 200
Q = np.array([[0.5, 0],
              [0, 0.5]])
R = np.array([[0.5, 0],
              [0, 0.5]])

b_des = (goal_pos,
         np.array([[1e-6, 0.0],
                   [0.0, 1e-6]]))
# u_des = [0.5*np.random.rand(2) for _ in range(num_segments)]
u_des = [[0.0, 0.0] for _ in range(num_segments)]

blqr = pomdp_py.BLQR(func_sysd, func_obs, jac_sysd, jac_obs, jac_sysd_u,
                     noise_obs, noise_sysd,
                     None, L, Q, R,
                     planning_horizon=planning_horizon)


def manual_test(blqr):
    # Initial and final belief states.
    b_0 = (np.array([2.0, 2.0]),
           np.array([[5.0, 0.0],
                     [0.0, 5.0]]))
    print("Path through the light:")
    print(b_0[0])
    print(b_0[1])
    with np.printoptions(precision=3, suppress=True):
        b_1 = blqr.ekf_update_mlo(b_0, [1.0, 0.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [1.0, 0.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [0.0, -1.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [0.0, -1.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [-1.0, 0.0])
        print(b_1[0])
        print(b_1[1])    
        b_1 = blqr.ekf_update_mlo(b_1, [-1.0, 0.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [-1.0, 0.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [-1.0, 0.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [-0.99, 0.09])
        print(b_1[0])
        print(b_1[1])


    # Initial and final belief states.
    b_0 = (np.array([2.0, 2.0]),
           np.array([[5.0, 0.0],
                     [0.0, 5.0]]))
    print("Path directly to goal")
    print(b_0[0])
    print(b_0[1])
    with np.printoptions(precision=3, suppress=True):
        b_1 = blqr.ekf_update_mlo(b_0, [-1.0, -1.0])
        print(b_1[0])
        print(b_1[1])
        b_1 = blqr.ekf_update_mlo(b_1, [-0.99, -0.99])
        print(b_1[0])
        print(b_1[1])

    ####### FOR DEBUGGING
    # traj through light
    u_traj_light = [[1.0, 0.0],
                    [1.0, 0.0],
                    [0.0, -1.0],
                    [0.0, -1.0],
                    [-1.0, 0.0],
                    [-1.0, 0.0],
                    [-1.0, 0.0],
                    [-1.0, 0.0],
                    [-0.99, 0.09]]
    b_t = b_0
    b_traj_light = [b_t]
    for t in range(len(u_traj_light)):
        u_t = u_traj_light[t]
        b_tp1 = blqr.ekf_update_mlo(b_t, u_t)
        b_traj_light.append(b_tp1)
        b_t = b_tp1
    bu_traj_light = [(b_traj_light[t], np.array(u_traj_light[t])) for t in range(len(u_traj_light))]
    
    # traj not through light
    u_traj_dark = [[-1., -1.], [-1., -1.], [-1., -1.]]
    b_t = b_0
    b_traj_dark = [b_t]
    for t in range(len(u_traj_dark)):
        u_t = u_traj_dark[t]
        b_tp1 = blqr.ekf_update_mlo(b_t, u_t)
        b_traj_dark.append(b_tp1)
        b_t = b_tp1        
    bu_traj_dark = [(b_traj_dark[t], np.array(u_traj_dark[t])) for t in range(len(u_traj_dark))]


    total_light = 0
    total_dark = 0    
    for i in range(1000):
        cost_light = blqr.segmented_cost_function(bu_traj_light, b_des, [], len(bu_traj_light))
        cost_dark = blqr.segmented_cost_function(bu_traj_dark, b_des, [], len(bu_traj_dark))
        total_light += cost_light
        total_dark += cost_dark
    print("avg cost light: %.3f" % (total_light/1000.0))
    print("avg cost dark: %.3f" % (total_dark/1000.0))

    x_range = (-1, 7)
    y_range = (-2, 4)
    viz = ld.LightDarkViz(env, x_range, y_range, 0.1)
    viz.set_goal(goal_pos)
    viz.set_initial_belief_pos(b_0[0])
    
    # Made up paths
    viz.log_position(tuple(b_0[0]), path=2)
    viz.log_position(tuple(b_0[0]), path=3)
    sysd_b_light = [b_0]
    for b_t, u_t in bu_traj_light:
        viz.log_position(tuple(b_t[0]), path=2)
        viz.log_position(tuple(sysd_b_light[-1][0]), path=3)
        sysd_b_light.append((func_sysd(sysd_b_light[-1][0], u_t), 0))

    # Made up paths
    viz.log_position(tuple(b_0[0]), path=4)
    viz.log_position(tuple(b_0[0]), path=5)
    sysd_b_dark = [b_0]
    for b_t, u_t in bu_traj_dark:
        viz.log_position(tuple(b_t[0]), path=4)
        viz.log_position(tuple(sysd_b_dark[-1][0]), path=5)
        sysd_b_dark.append((func_sysd(sysd_b_dark[-1][0], u_t), 0))        

    viz.plot(path_colors={2: [(0,0,0), (0,255,0)],
                          3: [(0,0,0), (0,255,255)],
                          4: [(0,0,0), (255,255,0)],
                          5: [(0,0,0), (255,0,255)]},
             path_styles={2: "--",
                          3: "-",
                          4: "--",
                          5: "-"})
    plt.show()    



def opt_callback(xk, *args):
    global count
    print("Iteration %d" % count)
    with np.printoptions(precision=3, suppress=True):
        print(xk)
    count += 1


def test(blqr):
    ############
    b_0 = (np.array([2.0, 2.0]),
           np.array([[5.0, 0.0],
                     [0.0, 5.0]]))    
    x_sol = blqr.create_plan(b_0, b_des, u_des,
                             num_segments=num_segments,
                             opt_options={"maxiter": 30},
                             opt_callback=opt_callback,
                             control_bounds=(-0.1, 0.1))
    with np.printoptions(precision=3, suppress=True):    
        print("SLSQP solution:")
        print(x_sol)
    plan = blqr.interpret_sqp_plan(x_sol, num_segments)

    # Visualization
    x_range = (-1, 7)
    y_range = (-2, 4)
    viz = ld.LightDarkViz(env, x_range, y_range, 0.1)
    viz.set_goal(goal_pos)
    viz.set_initial_belief_pos(b_0[0])
    viz.log_position(tuple(b_0[0]), path=0)
    viz.log_position(tuple(b_0[0]), path=1)

    sysd_b_plan = [b_0]
    for m_i, _, _ in plan:
        viz.log_position(tuple(m_i), path=0)

    viz.plot(path_colors={0: [(0,0,0), (0,255,0)],
                          1: [(0,0,0), (255,0,0)]},
             path_styles={0: "--",
                          1: "-"},
             path_widths={0: 4,
                          1: 1})
    plt.show()    

    


if __name__ == "__main__":
    test(blqr)
    # bt = (np.array([1.3, 2.0]), np.array([[0.5, 0.0], [0.0, 0.5]]))
    # ut = np.array([1.5, 0.2])
    
    # func_sysd = env.transition_model.func()
    # func_obs = obsmodel.func()
    # jac_sysd = env.transition_model.jac_dx()
    # jac_sysd_u = env.transition_model.jac_du()
    # jac_obs = obsmodel.jac_dx()
    
    # noise_sysd = pomdp_py.Gaussian([0,0], [[0.1,0],
    #                                        [0,0.1]]).random()
    # noise_obs_cov = obsmodel.noise_covariance(bt[0])

    # blqr = pomdp_py.BLQR(func_sysd, func_obs, jac_sysd, jac_obs, jac_sysd_u,
    #                      None, None, None, None)
    # start = time.time()
    # bnext = blqr.ekf_update_mlo(bt, ut, noise_sysd, noise_obs_cov)
    # print("Time taken:")
    # print(float(time.time() - start))
    # print("Mean:")
    # print(bnext[0])
    # print("Covariance:")
    # print(bnext[1])



"""
`pomdp_py <https://h2r.github.io/pomdp-py/html/>`_ provides function calls to use external solvers,
given a POMDP defined using pomdp_py interfaces. Currently, we interface with:

* `pomdp-solve <http://www.pomdp.org/code/index.html>`_ by Anthony R. Cassandra
* `SARSOP <https://github.com/AdaCompNUS/sarsop>`_ by NUS

We hope to interface with:

* `POMDP.jl <https://github.com/JuliaPOMDP/POMDPs.jl>`_
* more? Help us if you can!
"""
import pomdp_py
from pomdp_py.utils.interfaces.conversion\
    import to_pomdp_file, PolicyGraph, AlphaVectorPolicy, parse_pomdp_solve_output
import subprocess
import os, sys

def vi_pruning(agent, pomdp_solve_path,
               discount_factor=0.95,
               options=[],
               pomdp_name="temp-pomdp",
               remove_generated_files=False,
               return_policy_graph=False):
    """
    Value Iteration with pruning, using the software pomdp-solve
    https://www.pomdp.org/code/ developed by Anthony R. Cassandra.

    Args:
        agent (pomdp_py.Agent): The agent that contains the POMDP definition
        pomdp_solve_path (str): Path to the `pomdp_solve` binary generated after
            compiling the pomdp-solve library.
        options (list): Additional options to pass in to the command line interface.
             The options should be a list of strings, such as ["-stop_criteria", "weak", ...]
             Some useful options are:
                 -horizon <int>
                 -time_limit <int>
        pomdp_name (str): The name used to create the .pomdp file.
        remove_generated_files (bool): True if after policy is computed,
            the .pomdp, .alpha, .pg files are removed. Default is False.
        return_policy_graph (bool): True if return the policy as a PolicyGraph.
            By default is False, in which case an AlphaVectorPolicy is returned.

    Returns:
       PolicyGraph or AlphaVectorPolicy: The policy returned by the solver.
    """
    try:
        all_states = list(agent.all_states)
        all_actions = list(agent.all_actions)
        all_observations = list(agent.all_observations)
    except NotImplementedError:
        raise("S, A, O must be enumerable for a given agent to convert to .pomdp format")

    pomdp_path = "./%s.pomdp" % pomdp_name
    to_pomdp_file(agent, pomdp_path, discount_factor=discount_factor)
    proc = subprocess.Popen([pomdp_solve_path,
                             "-pomdp", pomdp_path,
                             "-o", pomdp_name] + list(map(str,options)))
    proc.wait()

    # Read the value and policy graph files
    alpha_path = "%s.alpha" % pomdp_name
    pg_path = "%s.pg" % pomdp_name
    if return_policy_graph:
        policy = PolicyGraph.construct(alpha_path, pg_path,
                                       all_states, all_actions, all_observations)
    else:
        policy = AlphaVectorPolicy.construct(
            alpha_path, all_states, all_actions, solver="pomdp-solve")

    # Remove temporary files
    if remove_generated_files:
        os.remove(pomdp_path)
        os.remove(alpha_path)
        os.remove(pg_path)
    return policy


def sarsop(agent, pomdpsol_path,
           discount_factor=0.95,
           timeout=30, memory=100,
           precision=0.5,
           pomdp_name="temp-pomdp",
           remove_generated_files=False,
           logfile=None):
    """
    SARSOP, using the binary from https://github.com/AdaCompNUS/sarsop
    This is an anytime POMDP planning algorithm

    Args:
        agent (pomdp_py.Agent): The agent that defines the POMDP models
        pomdpsol_path (str): Path to the `pomdpsol` binary
        timeout (int): The time limit (seconds) to run the algorithm until termination
        memory (int): The memory size (mb) to run the algorithm until termination
        precision (float): solver runs until regret is less than `precision`
        pomdp_name (str): Name of the .pomdp file that will be created when solving
        remove_generated_files (bool): Remove created files during solving after finish.
        logfile (str): Path to file to write the log of both stdout and stderr
    Returns:
       AlphaVectorPolicy: The policy returned by the solver.
    """
    try:
        all_states = list(agent.all_states)
        all_actions = list(agent.all_actions)
        all_observations = list(agent.all_observations)
    except NotImplementedError:
        raise("S, A, O must be enumerable for a given agent to convert to .pomdpx format")

    if logfile is None:
        stdout = None
        stderr = None
    else:
        logf = open(logfile, "w")
        stdout = subprocess.PIPE
        stderr = subprocess.STDOUT

    pomdp_path = "./%s.pomdp" % pomdp_name
    to_pomdp_file(agent, pomdp_path, discount_factor=discount_factor)
    proc = subprocess.Popen([pomdpsol_path,
                             "--timeout", str(timeout),
                             "--memory", str(memory),
                             "--precision", str(precision),
                             "--output", "%s.policy" % pomdp_name,
                             pomdp_path], stdout=stdout, stderr=stderr)
    if logfile is not None:
        for line in proc.stdout:
            line = line.decode("utf-8")
            sys.stdout.write(line)
            logf.write(line)
    proc.wait()

    policy_path = "%s.policy" % pomdp_name
    policy = AlphaVectorPolicy.construct(policy_path,
                                         all_states, all_actions)

    # Remove temporary files
    if remove_generated_files:
        os.remove(pomdp_path)
        os.remove(policy_path)
    if logfile is not None:
        logf.close()
    return policy



import pomdp_py
from pomdp_problems.tag.domain.state import *
from pomdp_problems.tag.models.transition_model import *
from pomdp_problems.tag.models.reward_model import *
from pomdp_problems.tag.models.components.motion_policy import *
from pomdp_problems.tag.models.components.grid_map import *
from pomdp_problems.multi_object_search.env.env import interpret
from pomdp_problems.multi_object_search.env.visual import MosViz

class TagEnvironment(pomdp_py.Environment):

    def __init__(self,
                 init_state,
                 grid_map,
                 pr_stay=0.2,
                 small=1,
                 big=10):
        self._grid_map = grid_map
        target_motion_policy = TagTargetMotionPolicy(grid_map,
                                                     pr_stay)
        transition_model = TagTransitionModel(grid_map, target_motion_policy)
        reward_model = TagRewardModel(small=small, big=big)
        super().__init__(init_state,
                         transition_model,
                         reward_model)

    @property
    def width(self):
        return self._grid_map.width
    
    @property
    def length(self):
        return self._grid_map.length

    @property
    def grid_map(self):
        return self._grid_map

    @classmethod
    def from_str(cls, worldstr, **kwargs):
        dim, robots, objects, obstacles, _ = interpret(worldstr)
        assert len(robots) == 1, "Does not support multiple robots."
        robot_position = robots[list(robots.keys())[0]].pose[:2]
        targets = []
        obstacle_poses = set({})
        for objid in objects:
            if objid not in obstacles:
                targets.append(objid)
            else:
                obstacle_poses.add(objects[objid].pose)
        assert len(targets) == 1, "Does not support multiple objects."                        
        target_position = objects[targets[0]].pose
        init_state = TagState(robot_position, target_position, False)
        grid_map = GridMap(dim[0], dim[1], obstacle_poses)
        return TagEnvironment(init_state, grid_map, **kwargs)



"""This module contains utility functions making it easier to debug POMDP
planning.

TreeDebugger
************

The core debugging functionality for POMCP/POUCT search trees is incorporated
into the TreeDebugger.  It is designed for ease of use during a :code:`pdb` or
:code:`ipdb` debugging session. Here is a minimal example usage:

.. code-block:: python

   from pomdp_py.utils import TreeDebugger
   from pomdp_problems.tiger import TigerProblem

   # pomdp_py.Agent
   agent = TigerProblem.create("tiger-left", 0.5, 0.15).agent

   # suppose pouct is a pomdp_py.POUCT object (POMCP works too)
   pouct = pomdp_py.POUCT(max_depth=4, discount_factor=0.95,
                          num_sims=4096, exploration_const=200,
                          rollout_policy=tiger_problem.agent.policy_model)

   action = pouct.plan(agent)
   dd = TreeDebugger(agent.tree)
   import pdb; pdb.set_trace()

When the program executes, you enter the pdb debugger, and you can:

.. code-block:: text

    (Pdb) dd.pp
    _VNodePP(n=4095, v=-19.529)(depth=0)
    ├─── ₀listen⟶_QNodePP(n=4059, v=-19.529)
    │    ├─── ₀tiger-left⟶_VNodePP(n=2013, v=-16.586)(depth=1)
    │    │    ├─── ₀listen⟶_QNodePP(n=1883, v=-16.586)
    │    │    │    ├─── ₀tiger-left⟶_VNodePP(n=1441, v=-8.300)(depth=2)
    ... # prints out the entire tree; Colored in terminal.

    (Pdb) dd.p(1)
    _VNodePP(n=4095, v=-19.529)(depth=0)
    ├─── ₀listen⟶_QNodePP(n=4059, v=-19.529)
    │    ├─── ₀tiger-left⟶_VNodePP(n=2013, v=-16.586)(depth=1)
    │    │    ├─── ₀listen⟶_QNodePP(n=1883, v=-16.586)
    │    │    ├─── ₁open-left⟶_QNodePP(n=18, v=-139.847)
    │    │    └─── ₂open-right⟶_QNodePP(n=112, v=-57.191)
    ... # prints up to depth 1

Note that the printed texts are colored in the terminal.

You can retrieve the subtree through indexing:

.. code-block:: text

    (Pdb) dd[0]
    listen⟶_QNodePP(n=4059, v=-19.529)
        - [0] tiger-left: VNode(n=2013, v=-16.586)
        - [1] tiger-right: VNode(n=2044, v=-16.160)

    (Pdb) dd[0][1][2]
    open-right⟶_QNodePP(n=15, v=-148.634)
        - [0] tiger-left: VNode(n=7, v=-20.237)
        - [1] tiger-right: VNode(n=6, v=8.500)

You can obtain the currently preferred action sequence by:

.. code-block:: text

    (Pdb) dd.mbp
       listen  []
       listen  []
       listen  []
       listen  []
       open-left  []
     _VNodePP(n=4095, v=-19.529)(depth=0)
     ├─── ₀listen⟶_QNodePP(n=4059, v=-19.529)
     │    └─── ₁tiger-right⟶_VNodePP(n=2044, v=-16.160)(depth=1)
     │         ├─── ₀listen⟶_QNodePP(n=1955, v=-16.160)
     │         │    └─── ₁tiger-right⟶_VNodePP(n=1441, v=-8.300)(depth=2)
     │         │         ├─── ₀listen⟶_QNodePP(n=947, v=-8.300)
     │         │         │    └─── ₁tiger-right⟶_VNodePP(n=768, v=0.022)(depth=3)
     │         │         │         ├─── ₀listen⟶_QNodePP(n=462, v=0.022)
     │         │         │         │    └─── ₁tiger-right⟶_VNodePP(n=395, v=10.000)(depth=4)
     │         │         │         │         ├─── ₁open-left⟶_QNodePP(n=247, v=10.000)

:code:`mbp` stands for "mark best plan".

To explore more features, browse the list of methods in the documentation.
"""
import sys
from pomdp_py.algorithms.po_uct import TreeNode, QNode, VNode, RootVNode
from pomdp_py.utils import typ, similar, special_char

SIMILAR_THRESH = 0.6
DEFAULT_MARK_COLOR = "blue"
MARKED = {}  # tracks marked nodes on tree

def _node_pp(node, e=None, p=None, o=None):
    # We want to return the node, but we don't want to print it on pdb with
    # its default string. But instead, we want to print it with our own
    # string formatting.
    if isinstance(node, VNode):
        return _VNodePP(node, parent_edge=e, parent=p, original=o)
    else:
        return _QNodePP(node, parent_edge=e, parent=p, original=o)

class _NodePP:

    def __init__(self, node, parent_edge=None, parent=None, original=None):
        """node: either VNode or QNode (the actual node on the tree) """
        self.parent_edge = parent_edge
        self.parent = parent
        self.children = node.children
        self.print_children = True
        if original is None:
            self.original = node
        else:
            self.original = original

    def __hash__(self):
        return id(self.original)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return id(self.original) == id(other.original)
        else:
            return False

    @property
    def marked(self):
        return id(self.original) in MARKED

    def to_edge(self, key):
        if key in self.children:
            return key
        elif type(key) == int:
            edges = list(sorted_by_str(self.children.keys()))
            return edges[key]
        elif type(key) == str:
            chosen = max(self.children.keys(),
                         key=lambda edge: similar(str(edge), key))
            if similar(str(chosen), key) >= SIMILAR_THRESH:
                return chosen
        raise ValueError("Cannot access children with key {}".format(key))

    def __getitem__(self, key):
        """
        When debugging, you can access the child of a node by the key
        of the following types:
        - the key is an action or observation object that points to a child;
          that is, key in self.children is True.
        - the key is an integer corresponding to the list of children shown
          when printing the node in the debugger
        - the key is a string that is similar to the string
          version of any of the action or observation edges;
          the most similar one will be chosen; The threshold
          of similarity is SIMILAR_THRESH
        """
        edge = self.to_edge(key)
        c = self.children[edge]
        if isinstance(c, _NodePP):
            original = c.original
        else:
            original = None
        return _node_pp(c, e=edge, p=self, o=original)

    def __contains__(self, key):
        try:
            self.to_edge(key)
            return True
        except ValueError:
            return False

    @staticmethod
    def interpret_print_type(opt):
        if opt.startswith("b") or opt.startswith("m"):
            opt = "marked-only"
        elif opt.startswith("s"):
            opt = "summary"
        elif opt.startswith("c"):
            opt = "complete"
        else:
            raise ValueError("Cannot understand print type: {}".format(opt))
        return opt

    def p(self, opt=None, **kwargs):
        if opt is None:
            max_depth = None
            print_type_opt = kwargs.get('t', "summary")
        elif type(opt) == int:
            max_depth = opt
            print_type_opt = kwargs.get('t', "summary")
        elif type(opt) == str:
            print_type_opt = opt
            max_depth = kwargs.get('d', None)
        else:
            raise ValueError("Cannot deal with opt of type {}".format(type(opt)))
        self.print_tree(max_depth=max_depth,
                        print_type=_NodePP.interpret_print_type(print_type_opt))

    @property
    def pp(self):
        self.print_tree(max_depth=None)

    def print_tree(self, **options):
        """Prints the tree, rooted at self"""
        _NodePP._print_tree_helper(self, 0, "", [None], -1, **options)

    @staticmethod
    def _print_tree_helper(root,
                           depth,   # depth of root
                           parent_edge,
                           branch_positions,  # list of 'first', 'middle', 'last' for each level prior to root
                           child_index,  # Index of the root as a child of parent
                           max_depth=None,
                           print_type="summary"):
        """
        pos_among_children is either 'first', 'middle', or 'last'
        """
        if max_depth is not None and depth > max_depth:
            return
        if root is None:
            return

        # Print the tree branches for all levels up to current root
        branches = ""
        preceding_positions = branch_positions[:-1]  # all positions except for current root
        for pos in preceding_positions:
            if pos is None:
                continue
            elif pos == "first" or pos == "middle":
                branches += "│    "
            else:  # "last"
                branches += "     "

        last_position = branch_positions[-1]
        if last_position is None:
            pass
        elif last_position == "first" or last_position == "middle":
            branches += "├─── "
        else:  # last
            branches += "└─── "

        root.print_children = False
        if child_index >= 0:
            line = branches + str(child_index).translate(special_char.SUBSCRIPT) + str(root)
        else:
            line = branches + str(root)
        if isinstance(root, VNode):
            line += typ.cyan("(depth="+str(depth)+")")

        print(line)

        for i, c in enumerate(sorted_by_str(root.children)):

            skip = True
            if root[c].marked:
                skip = False
            elif print_type == "complete":
                skip = False
            elif (root[c].num_visits > 1):
                skip = False
            if print_type == "marked-only" and not root[c].marked:
                skip = True

            if not skip:
                if isinstance(root[c], QNode):
                    next_depth = depth
                else:
                    next_depth = depth + 1

                if i == len(root.children) - 1:
                    next_pos = "last"
                elif i == 0:
                    next_pos = "first"
                else:
                    next_pos = "middle"

                _NodePP._print_tree_helper(root[c],
                                           next_depth,
                                           c,
                                           branch_positions + [next_pos],
                                           i,
                                           max_depth=max_depth,
                                           print_type=print_type)


class _QNodePP(_NodePP, QNode):
    """QNode for better printing"""
    def __init__(self, qnode, **kwargs):
        QNode.__init__(self, qnode.num_visits, qnode.value)
        _NodePP.__init__(self, qnode, **kwargs)

    def __str__(self):
        return TreeDebugger.single_node_str(self,
                                            parent_edge=self.parent_edge,
                                            include_children=self.print_children)

class _VNodePP(_NodePP, VNode):
    """VNode for better printing"""
    def __init__(self, vnode, **kwargs):
        VNode.__init__(self, vnode.num_visits)
        _NodePP.__init__(self, vnode, **kwargs)

    def __str__(self):
        return TreeDebugger.single_node_str(self,
                                            parent_edge=self.parent_edge,
                                            include_children=self.print_children)


class TreeDebugger:
    """
    Helps you debug the search tree; A search tree is a tree
    that contains a subset of future histories, organized into
    QNodes (value represents Q(b,a); children are observations) and
    VNodes (value represents V(b); children are actions).
    """
    def __init__(self, tree):
        """
        Args:
            tree (VNode): the root node of a search tree. For example,
                the tree built by POUCT after planning an action,
                which can be accessed by agent.tree.
        """
        if not isinstance(tree, TreeNode):
            raise ValueError("Expecting tree to be a TreeNode, but got {}".format(type(tree)))

        self.tree = _node_pp(tree)
        self.current = self.tree   # points to the node the user is interacting with
        self._stats_cache = {}

    def __str__(self):
        return str(self.current)

    def __repr__(self):
        nodestr = TreeDebugger.single_node_str(self.current,
                                               parent_edge=self.current.parent_edge)
        return "TreeDebugger@\n{}".format(nodestr)

    def __getitem__(self, key):
        if type(key) == tuple:
            n = self.current
            for k in key:
                n = n[k]
            return n
        else:
            return self.current[key]

    def _get_stats(self):
        if id(self.current) in self._stats_cache:
            stats = self._stats_cache[id(self.current)]
        else:
            stats = TreeDebugger.tree_stats(self.current)
            self._stats_cache[id(self.current)] = stats
        return stats

    def num_nodes(self, kind='all'):
        """
        Returns the total number of nodes in the tree rooted at "current"
        """
        stats = self._get_stats()
        res = {
            'all': stats['total_vnodes'] + stats['total_qnodes'],
            'q': stats['total_qnodes'],
            'v': stats['total_vnodes']
        }
        if kind in res:
            return res[kind]
        else:
            raise ValueError("Invalid value for kind={}; Valid values are {}"\
                             .format(kind, list(res.keys())))


    @property
    def depth(self):
        """Tree depth starts from 0 (root node only).
        It is the largest number of edges on a path from root to leaf."""
        stats = self._get_stats()
        return stats['max_depth']

    @property
    def d(self):
        """alias for depth"""
        return self.depth

    @property
    def num_layers(self):
        """Returns the number of layers;
        It is the number of layers of nodes, which equals to depth + 1"""
        return self.depth + 1

    @property
    def nl(self):
        """alias for num_layers"""
        return self.num_layers

    @property
    def nn(self):
        """Returns the total number of nodes in the tree"""
        return self.num_nodes(kind='all')

    @property
    def nq(self):
        """Returns the total number of QNodes in the tree"""
        return self.num_nodes(kind='q')

    @property
    def nv(self):
        """Returns the total number of VNodes in the tree"""
        return self.num_nodes(kind='v')

    def l(self, depth, as_debuggers=True):
        """alias for layer"""
        return self.layer(depth, as_debuggers=as_debuggers)

    def layer(self, depth, as_debuggers=True):
        """
        Returns a list of nodes at the given depth. Will only return VNodes.
        Warning: If depth is high, there will likely be a huge number of nodes.

        Args:
            depth (int): Depth of the tree
            as_debuggers (bool): True if return a list of TreeDebugger objects,
                one for each tree on the layer.
        """
        if depth < 0 or depth > self.depth:
            raise ValueError("Depth {} is out of range (0-{})".format(depth, self.depth))
        nodes = []
        self._layer_helper(self.current, 0, depth, nodes)
        return nodes

    def _layer_helper(self, root, current_depth, target_depth, nodes, as_debuggers=True):
        if current_depth == target_depth:
            if isinstance(root, VNode):
                if as_debuggers:
                    nodes.append(TreeDebugger(root.original))
                else:
                    nodes.append(root)
        else:
            for c in sorted_by_str(root.children):
                if isinstance(root[c], QNode):
                    next_depth = current_depth
                else:
                    next_depth = current_depth + 1
                self._layer_helper(root[c],
                                   next_depth,
                                   target_depth,
                                   nodes)

    @property
    def leaf(self):
        worklist = [self.current]
        seen = set({self.current})
        leafs = []
        while len(worklist) > 0:
            node = worklist.pop()
            if len(node.children) == 0:
                leafs.append(node)
            else:
                for c in node.children:
                    if node[c] not in seen:
                        worklist.append(node[c])
                        seen.add(node[c])
        return leafs

    def step(self, key):
        """Updates current interaction node to follow the
        edge along key"""
        edge = self.current.to_edge(key)
        self.current = self[edge]
        print("step: " + str(edge))


    def s(self, key):
        """alias for step"""
        return self.step(key)

    def back(self):
        """move current node of interaction back to parent"""
        self.current = self.current.parent

    @property
    def b(self):
        """alias for back"""
        self.back()

    @property
    def root(self):
        """The root node when first creating this TreeDebugger"""
        return self.tree

    @property
    def r(self):
        """alias for root"""
        return self.root

    @property
    def c(self):
        """Current node of interaction"""
        return self.current

    def p(self, *args, **kwargs):
        """print tree"""
        return self.current.p(*args, **kwargs)

    @property
    def pp(self):
        """print tree, with preset options"""
        return self.current.pp

    @property
    def mbp(self):
        """Mark Best and Print.
        Mark the best sequence, and then print with only the marked nodes"""
        self.mark(self.bestseq, color="yellow")
        self.p("marked-only")

    @property
    def pm(self):
        """Print marked only"""
        self.p("marked-only")

    def mark_sequence(self, seq, color=DEFAULT_MARK_COLOR):
        """
        Given a list of keys (understandable by __getitem__ in _NodePP),
        mark nodes (both QNode and VNode) along the path in the tree.
        Note this sequence starts from self.current; So self.current will
        also be marked.
        """
        node = self.current
        MARKED[id(node.original)] = interpret_color(color)
        for key in seq:
            MARKED[id(node[key].original)] = interpret_color(color)
            node = node[key]

    def mark(self, seq, **kwargs):
        """alias for mark_sequence"""
        return self.mark_sequence(seq, **kwargs)

    def mark_path(self, dest, **kwargs):
        """paths the path to dest node"""
        return self.mark(self.path_to(dest), **kwargs)

    def markp(self, dest, **kwargs):
        """alias to mark_path"""
        return self.mark_path(dest, **kwargs)

    @property
    def clear(self):
        """Clear the marks"""
        global MARKED
        MARKED = {}

    @property
    def bestseq(self):
        """Returns a list of actions, observation sequence
        that have the highest value for each step. Such
        a sequence is "preferred".

        Also, prints out the list of preferred actions for each step
        into the future"""
        return self.preferred_actions(self.current, max_depth=None)

    def bestseqd(self, max_depth):
        """
        alias for bestseq except with
        """
        return self.preferred_actions(self.current, max_depth=max_depth)

    @staticmethod
    def single_node_str(node, parent_edge=None, indent=1, include_children=True):
        """
        Returns a string for printing given a single vnode.
        """
        if hasattr(node, "marked") and node.marked:
            color_fn = MARKED[id(node.original)]
            opposite_color = color = lambda s: typ.bold(color_fn(s))
        elif isinstance(node, VNode):
            color = typ.green
            opposite_color = typ.red
        else:
            assert isinstance(node, QNode)
            color = typ.red
            opposite_color = typ.green

        output = ""
        if parent_edge is not None:
            output += opposite_color(str(parent_edge)) + "⟶"

        output += color(str(node.__class__.__name__))\
            + "(n={}, v={:.3f})".format(node.num_visits,
                                    node.value)
        if include_children:
            output += "\n"
            for i, action in enumerate(sorted_by_str(node.children)):
                child = node.children[action]
                child_info = TreeDebugger.single_node_str(child, include_children=False)

                spaces = "    " * indent
                output += "{}- [{}] {}: {}".format(spaces, i,
                                                   typ.white(str(action)),
                                                   child_info)
                if i < len(node.children) - 1:
                    output += "\n"
            output += "\n"
        return output

    @staticmethod
    def preferred_actions(root, max_depth=None):
        """
        Print out the currently preferred actions up to given `max_depth`
        """
        seq = []
        TreeDebugger._preferred_actions_helper(root, 0, seq, max_depth=max_depth)
        return seq

    @staticmethod
    def _preferred_actions_helper(root, depth, seq, max_depth=None):
        # don't care about last layer action because it's outside of planning
        # horizon and only has initial value.
        if max_depth is not None and depth > max_depth:
            return
        if root is None or len(root.children) == 0:
            return
        best_child = root.to_edge(0)
        best_value = root[0].value
        for c in root.children:
            if root[c].value > best_value:
                best_child = c
                best_value = root[c].value
        seq.append(best_child)
        equally_good = []
        if isinstance(root, VNode):
            for c in root.children:
                if not(c == best_child) and root[c].value == best_value:
                    equally_good.append(c)

        if best_child is not None and root[best_child] is not None:
            if isinstance(root[best_child], QNode):
                print("  %s  %s" % (typ.yellow(str(best_child)), str(equally_good)))
                next_depth = depth
            else:
                next_depth = depth + 1

            TreeDebugger._preferred_actions_helper(root[best_child], next_depth, seq,
                                                   max_depth=max_depth)

    def path(self, dest):
        """alias for path_to;
        Example usage:

        marking path from root to the first node on the second layer:

            dd.mark(dd.path(dd.layer(2)[0]))
        """
        return self.path_to(dest)

    def path_to(self, dest):
        """Returns a list of keys (actions / observations) that represents the path from
        self.current to the given node `dest`. Returns None if the path does not
        exist.  Uses DFS. Can be useful for marking path to a node to a specific
        layer. Note that the returned path is a list of keys (i.e. edges), not nodes.
        """
        # dest may be in the returned list of layer() which could be a TreeDebugger.
        if isinstance(dest, TreeDebugger):
            dest = dest.current
        worklist = [self.current]
        seen = set({self.current})
        parent = {self.current: None}
        while len(worklist) > 0:
            node = worklist.pop()
            if node == dest:
                return self._get_path(self.current, dest, parent)
            for c in node.children:
                if node[c] not in seen:
                    worklist.append(node[c])
                    seen.add(node[c])
                    parent[node[c]] = (node, c)
        return None

    def _get_path(self, start, dest, parent):
        """Helper method for path_to"""
        v = dest
        path = []
        while v != start:
            v, edge = parent[v]
            path.append(edge)
        return list(reversed(path))

    @staticmethod
    def tree_stats(root, max_depth=None):
        """Gether statistics about the tree"""
        stats = {
            'total_vnodes': 0,
            'total_qnodes': 0,
            'total_vnodes_children': 0,
            'total_qnodes_children': 0,
            'max_vnodes_children': 0,
            'max_qnodes_children': 0,
            'max_depth': 0
        }
        TreeDebugger._tree_stats_helper(root, 0, stats, max_depth=max_depth)
        stats['num_visits'] = root.num_visits
        stats['value'] = root.value
        return stats

    @staticmethod
    def _tree_stats_helper(root, depth, stats, max_depth=None):
        if max_depth is not None and depth > max_depth:
            return
        else:
            if isinstance(root, VNode):
                stats['total_vnodes'] += 1
                stats['total_vnodes_children'] += len(root.children)
                stats['max_vnodes_children'] = max(stats['max_vnodes_children'], len(root.children))
                stats['max_depth'] = max(stats['max_depth'], depth)
            else:
                stats['total_qnodes'] += 1
                stats['total_qnodes_children'] += len(root.children)
                stats['max_qnodes_children'] = max(stats['max_qnodes_children'], len(root.children))

            for c in root.children:
                if isinstance(root[c], QNode):
                    next_depth = depth
                else:
                    next_depth = depth + 1
                TreeDebugger._tree_stats_helper(root[c], next_depth, stats, max_depth=max_depth)

def sorted_by_str(enumerable):
    return sorted(enumerable, key=lambda n: str(n))

def interpret_color(colorstr):
    if colorstr.lower() in typ.colors:
        return eval("typ.{}".format(colorstr))
    else:
        raise ValueError("Invalid color: {};\n"
                         "The available ones are {}".format(colorstr, typ.colors))



import pomdp_py
import copy
import random
from pomdp_problems.tag.domain.observation import *
from pomdp_problems.tag.domain.action import *
from pomdp_problems.tag.domain.state import *
from pomdp_problems.tag.models.observation_model import *
from pomdp_problems.tag.models.transition_model import *
from pomdp_problems.tag.models.reward_model import *
from pomdp_problems.tag.models.policy_model import *
from pomdp_problems.tag.models.components.motion_policy import *
from pomdp_problems.tag.models.components.grid_map import *

## initialize belief
def initialize_belief(grid_map, init_robot_position, prior={}):
    """Initialize belief.
    
    Args:
        grid_map (GridMap): Holds information of the map occupancy
        prior (dict): A map from (x,y)->[0,1]. If empty, the belief
            will be uniform."""
    hist = {}  # state -> prob
    total_prob = 0.0
    for target_position in prior:
        state = TagState(init_robot_position, target_position, False)
        hist[state] = prior[target_position]
        total_prob += hist[state]

    for x in range(grid_map.width):
        for y in range(grid_map.length):
            if (x,y) in grid_map.obstacle_poses:
                # Skip obstacles
                continue
            state = TagState(init_robot_position, (x,y), False)
            if len(prior) > 0:
                if (x,y) not in prior:
                    hist[state] = 1e-9
            else:
                hist[state] = 1.0
                total_prob += hist[state]
    # Normalize
    for state in hist:
        hist[state] /= total_prob

    hist_belief = pomdp_py.Histogram(hist)
    return hist_belief

def initialize_particles_belief(grid_map, init_robot_position, num_particles=100, prior={}):
    """Initialize belief.
    
    Args:
        grid_map (GridMap): Holds information of the map occupancy
        prior (dict): A map from (x,y)->[0,1]. If empty, the belief
            will be uniform."""
    particles = []
    if len(prior) > 0:
        # prior knowledge provided. Just use the prior knowledge
        prior_sum = sum(prior[pose] for pose in prior)
        for pose in prior[objid]:
            state = TagState(init_robot_position, pose)
            amount_to_add = (prior[objid][pose] / prior_sum) * num_particles
            for _ in range(amount_to_add):
                particles.append(state)    
    else:
        while len(particles) < num_particles:
            target_position = (random.randint(0, grid_map.width-1),
                               random.randint(0, grid_map.length-1))
            if target_position in grid_map.obstacle_poses:
                # Skip obstacles
                continue            
            state = TagState(init_robot_position, target_position, False)
            particles.append(state)
    return pomdp_py.Particles(particles)


## belief update
def belief_update(agent, real_action, real_observation):
    # Update agent belief
    current_mpe_state = agent.cur_belief.mpe()
    next_robot_position = agent.transition_model.sample(current_mpe_state, real_action).robot_position
    
    next_state_space = set({})
    for state in agent.cur_belief:
        next_state = copy.deepcopy(state)
        next_state.robot_position = next_robot_position
        next_state_space.add(next_state)

    new_belief = pomdp_py.update_histogram_belief(
        agent.cur_belief, real_action, real_observation,
        agent.observation_model, agent.transition_model,
        next_state_space=next_state_space)
    
    agent.set_belief(new_belief)

class TagAgent(pomdp_py.Agent):
    
    def __init__(self,
                 init_belief,
                 grid_map,
                 pr_stay=0.2,
                 small=1,
                 big=10):
        self._grid_map = grid_map        
        target_motion_policy = TagTargetMotionPolicy(grid_map,
                                                     pr_stay)        
        transition_model = TagTransitionModel(grid_map,
                                              target_motion_policy)
        reward_model = TagRewardModel(small=small, big=big)
        observation_model = TagObservationModel()
        policy_model = TagPolicyModel(grid_map=grid_map)
        super().__init__(init_belief,
                         policy_model,
                         transition_model=transition_model,
                         observation_model=observation_model,
                         reward_model=reward_model)

    def clear_history(self):
        """Custum function; clear history"""
        self._history = None
        
        




import pomdp_py


# we index the walls around a grid cell in
# clockwise fashion: top wall (0), right wall (1),
# bottom wall (2), left wall (3).
WALL = {
    0: "top",
    1: "right",
    2: "bottom",
    3: "left"
}

class Observation(pomdp_py.Observation):
    def __init__(self, walls, orientation):
        """
        Args:
            walls (tuple) is a tuple of integers, that indicate the walls
            around a grid cell that are present.
            orientation (float) is the orientation of the robot.
        """
        self.walls = walls
        self.orientation = orientation

    def __hash__(self):
        return hash((self.wall_case, self.orientation))

    def __eq__(self, other):
        if not isinstance(other, Observation):
            return False
        else:
            return self.walls == other.walls\
                and self.orientation == other.orientation



import pomdp_py
from pomdp_problems.tag.domain.observation import *
import pomdp_problems.tag.constants as constants

class TagObservationModel(pomdp_py.ObservationModel):
    """In this observation model, the robot deterministically
    observes the target location when it is in the same grid cell
    as the target. Ohterwise the robot does not observe anything."""

    def probability(self, observation, next_state, action, **kwargs):
        if next_state.robot_position == next_state.target_position:
            if observation.target_position is None:
                return constants.EPSILON
            else:
                if observation.target_position == next_state.target_position:
                    return 1.0 - constants.EPSILON
                else:
                    return constants.EPSILON
        else:
            if observation.target_position is None:
                return 1.0 - constants.EPSILON
            else:
                return constants.EPSILON
    
    def sample(self, next_state, action):
        """There is no stochaisticity in the observation model"""
        if next_state.robot_position == next_state.target_position:
            return TagObservation(next_state.target_position)
        else:
            return TagObservation(None)

    def argmax(self, next_state, action):
        return self.sample(next_state, action)



"""Assorted utilities for math"""
import math
import numpy as np

# Math
def vec(p1, p2):
    """ vector from p1 to p2 """
    if type(p1) != np.ndarray:
        p1 = np.array(p1)
    if type(p2) != np.ndarray:
        p2 = np.array(p2)
    return p2 - p1

def proj(vec1, vec2, scalar=False):
    # Project vec1 onto vec2. Returns a vector in the direction of vec2.
    scale = np.dot(vec1, vec2) / np.linalg.norm(vec2)
    if scalar:
        return scale
    else:
        return vec2 * scale

def R_x(th):
    return np.array([
        1, 0, 0, 0,
        0, np.cos(th), -np.sin(th), 0,
        0, np.sin(th), np.cos(th), 0,
        0, 0, 0, 1
    ]).reshape(4,4)

def R_y(th):
    return np.array([
        np.cos(th), 0, np.sin(th), 0,
        0, 1, 0, 0,
        -np.sin(th), 0, np.cos(th), 0,
        0, 0, 0, 1
    ]).reshape(4,4)

def R_z(th):
    return np.array([
        np.cos(th), -np.sin(th), 0, 0,
        np.sin(th), np.cos(th), 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1
    ]).reshape(4,4)

def T(dx, dy, dz):
    return np.array([
        1, 0, 0, dx,
        0, 1, 0, dy,
        0, 0, 1, dz,
        0, 0, 0, 1
    ]).reshape(4,4)

def to_radians(th):
    return th*np.pi / 180

def R_between(v1, v2):
    if len(v1) != 3 or len(v2) != 3:
        raise ValueError("Only applicable to 3D vectors!")
    v = np.cross(v1, v2)
    c = np.dot(v1, v2)
    s = np.linalg.norm(v)
    I = np.identity(3)

    vX = np.array([
        0, -v[2], v[1],
        v[2], 0, -v[0],
        -v[1], v[0], 0
    ]).reshape(3,3)
    R = I + vX + np.matmul(vX,vX) * ((1-c)/(s**2))
    return R

def approx_equal(v1, v2, epsilon=1e-6):
    if len(v1) != len(v2):
        return False
    for i in range(len(v1)):
        if abs(v1[i] - v2[i]) > epsilon:
            return False
    return True

def euclidean_dist(p1, p2):
    return math.sqrt(sum([(a - b)** 2 for a, b in zip(p1, p2)]))



"""Implementation of the basic policy tree based value iteration as explained
in section 4.1 of `Planning and acting in partially observable stochastic
domains` :cite:`kaelbling1998planning`

Warning: No pruning - the number of policy trees explodes very fast.
"""

from pomdp_py.framework.planner cimport Planner
from pomdp_py.framework.basics cimport Agent, Action, State
import numpy as np
import itertools

cdef class _PolicyTreeNode:

    cdef public Action action
    cdef public int depth
    cdef public dict children, values
    cdef Agent _agent
    cdef float _discount_factor

    def __init__(self, action, depth, agent,
                 discount_factor, children={}):
        self.action = action
        self.depth = depth
        self._agent = agent
        self.children = children
        self._discount_factor = discount_factor
        self.values = self._compute_values()  # s -> value

    def _compute_values(self):
        """
        Returns a dictionary {s -> value} that represents the values
        for the next actions.
        """
        actions = self._agent.all_actions
        observations = self._agent.all_observations
        states = self._agent.all_states

        discount_factor = self._discount_factor**self.depth
        values = {}
        for s in states:
            expected_future_value = 0.0
            for sp in states:
                for o in observations:
                    trans_prob = self._agent.transition_model.probability(sp, s, self.action)
                    obsrv_prob = self._agent.observation_model.probability(o, sp, self.action)
                    if len(self.children) > 0:
                        subtree_value = self.children[o].values[s]  # corresponds to V_{oi(p)} in paper
                    else:
                        subtree_value = 0.0
                    reward = self._agent.reward_model.sample(s, self.action, sp)
                    expected_future_value += trans_prob * obsrv_prob * (reward + discount_factor*subtree_value)
            values[s] = expected_future_value
        return values

    def __str__(self):
        return "_PolicyTreeNode(%s, %d){%s}" % (self.action, self.depth, str(self.children_keys))
    def __repr__(self):
        return self.__str__()


cdef class ValueIteration(Planner):
    """
    This algorithm is only feasible for small problems where states, actions,
    and observations can be explicitly enumerated.

    __init__(self, horizon=float('inf'), discount_factor=0.9, epsilon=1e-6)
    """
    cdef float _discount_factor, _epsilon
    cdef int _planning_horizon

    def __init__(self, horizon, discount_factor=0.9, epsilon=1e-6):
        """
        The horizon satisfies discount_factor**horizon > epsilon"""
        assert type(horizon) == int and horizon >= 1, "Horizon must be an integer >= 1"
        self._discount_factor = discount_factor
        self._epsilon = epsilon
        self._planning_horizon = horizon

    def plan(self, agent):
        """plan(self, agent)
        Plans an action."""
        policy_trees = self._build_policy_trees(1, agent)
        value_beliefs = {}
        for p, policy_tree in enumerate(policy_trees):
            value_beliefs[p] = 0
            for state in agent.all_states:
                value_beliefs[p] += agent.cur_belief[state] * policy_tree.values[state]
        # Pick the policy tree with highest belief value
        pmax = max(value_beliefs, key=value_beliefs.get)
        return policy_trees[pmax].action

    def _build_policy_trees(self, depth, agent):
        """Bottom up build policy trees"""
        actions = agent.all_actions
        states = agent.all_states
        observations = agent.all_observations  # we expect observations to be indexed

        if depth >= self._planning_horizon or self._discount_factor**depth < self._epsilon:
            return [_PolicyTreeNode(a, depth, agent, self._discount_factor)
                    for a in actions]
        else:
            # Every observation can lead to K possible sub policy trees, which
            # is exactly the output of _build_policy_trees. Then, for a set of
            # observations, a policy tree is formed by combining together one
            # sub policy tree (corresponding to one action in the next time
            # step) per observation from the pool of K possible sub policy
            # trees. So we take the cartesian product of these sets of sub
            # policy trees and build individual policy trees.
            groups = [self._build_policy_trees(depth+1, agent)
                      for i in range(len(observations))]
            # (Sanity check) We expect all groups to have same size
            group_size = len(groups[0])
            for g in groups:
                assert group_size == len(g)

            # This computes all combinations of sub policy trees. Each combination
            # will become one policy tree that will be returned, with an action to
            # take at the current depth level as the root.
            combinations = itertools.product(*([np.arange(group_size)]*len(observations)))
            policy_trees = []
            for comb in combinations:
                # comb is a tuple of indicies, e.g. (i, j, k) that means
                # for observation 0, the sub policy tree is at index i of its group;
                # for observation 1, the sub policy tree is at index j of its group, etc.
                # We want to create a mapping from observation to sub policy tree.
                assert len(comb) == len(observations)  # sanity check
                children = {}
                for oi in range(len(comb)):
                    sub_policy_tree = groups[oi][comb[oi]]
                    children[observations[oi]] = sub_policy_tree

                # Now that we have the children, we know that there could be
                # |A| different root nodes, resulting in |A| different policy trees
                for a in actions:
                    policy_tree_node = _PolicyTreeNode(
                        a, depth, agent, self._discount_factor, children=children)
                    policy_trees.append(policy_tree_node)

            return policy_trees



#!/usr/bin/env python

from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import os.path


with open("README.rst", 'r') as f:
    long_description = f.read()

# Build cython files as extensions
def build_extensions(pkg_name, major_submodules):
    cwd = os.path.abspath(os.path.dirname(__file__))
    extensions = []
    for subm in major_submodules:
        for f in os.listdir(os.path.join(cwd, pkg_name, subm.replace(".", "/"))):
            if f.endswith(".pyx"):
                filename = os.path.splitext(f)[0]
                ext_name = f"{pkg_name}.{subm}.{filename}"
                ext_path = os.path.join(pkg_name, subm.replace(".", "/"), f)
                extensions.append(Extension(ext_name, [ext_path]))

    return extensions

extensions = build_extensions("pomdp_py", ["framework",
                                           "algorithms",
                                           "utils",
                                           "representations.distribution",
                                           "representations.belief"])
extensions += [
    Extension("pomdp_problems.tiger.cythonize", ["pomdp_problems/tiger/cythonize/tiger_problem.pyx"]),
    Extension("pomdp_problems.rocksample.cythonize", ["pomdp_problems/rocksample/cythonize/rocksample_problem.pyx"])
]

setup(name='pomdp-py',
      packages=find_packages(),
      version='1.3.1',
      description='Python POMDP Library.',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      install_requires=[
          'Cython',
          'numpy',
          'scipy',
          'tqdm',
          'matplotlib',
          'pygame',        # for some tests
          'opencv-python',  # for some tests
      ],
      license="MIT",
      author='Kaiyu Zheng',
      author_email='kzheng10@cs.brown.edu',
      keywords = ['Partially Observable Markov Decision Process', 'POMDP'],
      ext_modules=cythonize(extensions,
                            build_dir="build",
                            compiler_directives={'language_level' : "3"}),
      package_data={"pomdp_py": ["*.pxd", "*.pyx"],
                    "pomdp_problems": ["*.pxd", "*.pyx"]},
      zip_safe=False
)



import pomdp_py
import random
from pomdp_py.utils.math import euclidean_dist
import pomdp_problems.tag.constants as constants
from pomdp_problems.tag.models.transition_model import TagTransitionModel


class TagTargetMotionPolicy(pomdp_py.GenerativeDistribution):
    def __init__(self,
                 grid_map,
                 pr_stay=0.2):  # With 1.0 - pr_stay chance, the target moves away
        self._grid_map = grid_map
        self._pr_stay = pr_stay

    def _compute_candidate_actions(self,
                                   robot_position,
                                   target_position,
                                   valid_target_motion_actions):
        candidate_actions = set({})
        cur_dist = euclidean_dist(robot_position, target_position)
        for action in valid_target_motion_actions:
            next_target_position = TagTransitionModel.if_move_by(self._grid_map,
                                                                 target_position,
                                                                 action)
            next_dist = euclidean_dist(robot_position, next_target_position)
            if next_dist > cur_dist:
                candidate_actions.add(action)
        return candidate_actions


    def probability(self, next_target_position,
                    target_position, robot_position,
                    valid_target_motion_actions):
        # If it is impossible to go from target position to the next,
        # then it is a zero probability event.
        diff_x = abs(next_target_position[0] - target_position[0])
        diff_y = abs(next_target_position[1] - target_position[1])
        if not ((diff_x == 1 and diff_y == 0)
                or (diff_x == 0 and diff_y == 1)
                or (diff_x == 0 and diff_y == 0)):
            return constants.EPSILON

        candidate_actions = self._compute_candidate_actions(robot_position,
                                                            target_position,
                                                            valid_target_motion_actions)
        if len(candidate_actions) == 0:
            # No action possible, yet next_target_position is a valid
            # transition from current.
            if next_target_position == target_position:
                # That means the target is either
                # stuck or staying. Either way, this is the only thing that
                # can happen
                return 1.0 - constants.EPSILON
            else:
                return constants.EPSILON
        else:
            # There are candidate actions
            if next_target_position == target_position:
                # The object is staying
                return self._pr_stay
            else:
                # The object has taken an adversarial action.
                for action in candidate_actions:
                    if (target_position[0] + action.motion[0],
                        target_position[1] + action.motion[1]) == next_target_position:
                        return (1.0 - self._pr_stay) / len(candidate_actions)
                    return constants.EPSILON

    def random(self, robot_position, target_position, valid_target_motion_actions,
               mpe=False):
        if mpe or random.uniform(0,1) > self._pr_stay:
            # Move away; Pick motion actions that makes the target moves away from the robot
            candidate_actions = self._compute_candidate_actions(robot_position,
                                                                target_position,
                                                                valid_target_motion_actions)
            if len(candidate_actions) == 0:
                return target_position

            chosen_action = random.sample(candidate_actions, 1)[0]
            return TagTransitionModel.if_move_by(self._grid_map,
                                                 target_position,
                                                 chosen_action)
        else:
            # stay
            return target_position

    def mpe(self, robot_position, target_position, valid_target_motion_actions):
        return self.random(robot_position, target_position,
                           valid_target_motion_actions,
                           mpe=True)



"""Misc Python utilities"""
from difflib import SequenceMatcher

def remap(oldvalue, oldmin, oldmax, newmin, newmax):
    if oldmax - oldmin == 0:
        print("Warning in remap: the old range has size 0")
        oldmax = oldmin + oldvalue
    return (((oldvalue - oldmin) * (newmax - newmin)) / (oldmax - oldmin)) + newmin


def json_safe(obj):
    if isinstance(obj, bool):
        return str(obj).lower()
    elif isinstance(obj, (list, tuple)):
        return [json_safe(item) for item in obj]
    elif isinstance(obj, dict):
        return {json_safe(key):json_safe(value) for key, value in obj.items()}
    else:
        return str(obj)
    return obj


def safe_slice(arr, start, end):
    true_start = max(0, min(len(arr)-1, start))
    true_end = max(0, min(len(arr)-1, end))
    return arr[true_start:true_end]

def similar(a, b):
    # Reference: https://stackoverflow.com/questions/17388213/find-the-similarity-metric-between-two-strings
    return SequenceMatcher(None, a, b).ratio()


class special_char:
    left = '\u2190'
    up = '\u2191'
    right = '\u2192'
    down = '\u2193'
    longleft = '\u27F5'
    longright = '\u27F6'

    hline = "─"
    vline = "│"

    bottomleft = "└"
    longbottomleft = "└─"
    topleft = "┌"
    longtopleft = "┌─"
    topright = "┐"
    longtopright = "─┐"
    bottomright = "┘"
    longbottomright = "─┘"

    intersect = "┼"
    topt = "┬"
    leftt = "├"
    rightt = "┤"
    bottomt = "┴"
    shadebar = "▒"

    SUBSCRIPT = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")



"""Defines the ObservationModel for the 2D Multi-Object Search domain.

Origin: Multi-Object Search using Object-Oriented POMDPs (ICRA 2019)
(extensions: action space changes, different sensor model, gridworld instead of
topological graph)

Observation: {objid : pose(x,y) or NULL}. The sensor model could vary;
             it could be a fan-shaped model as the original paper, or
             it could be something else. But the resulting observation
             should be a map from object id to observed pose or NULL (not observed).

Observation Model

  The agent can observe its own state, as well as object poses
  that are within its sensor range. We only need to model object
  observation.

"""

import pomdp_py
import math
import random
import numpy as np
from pomdp_problems.multi_object_search.domain.state import *
from pomdp_problems.multi_object_search.domain.action import *
from pomdp_problems.multi_object_search.domain.observation import *

#### Observation Models ####
class MosObservationModel(pomdp_py.OOObservationModel):
    """Object-oriented transition model"""
    def __init__(self,
                 dim,
                 sensor,
                 object_ids,
                 sigma=0.01,
                 epsilon=1):
        self.sigma = sigma
        self.epsilon = epsilon
        observation_models = {objid: ObjectObservationModel(objid, sensor, dim,
                                                            sigma=sigma, epsilon=epsilon)
                              for objid in object_ids}
        pomdp_py.OOObservationModel.__init__(self, observation_models)

    def sample(self, next_state, action, argmax=False, **kwargs):
        if not isinstance(action, LookAction):
            return MosOOObservation({})
            # return MosOOObservation({objid: ObjectObservationModel.NULL
            #                          for objid in next_state.object_states
            #                          if objid != next_state.object_states[objid].objclass != "robot"})

        factored_observations = super().sample(next_state, action, argmax=argmax)
        return MosOOObservation.merge(factored_observations, next_state)

class ObjectObservationModel(pomdp_py.ObservationModel):
    def __init__(self, objid, sensor, dim, sigma=0, epsilon=1):
        """
        sigma and epsilon are parameters of the observation model (see paper),
        dim (tuple): a tuple (width, length) for the dimension of the world"""
        self._objid = objid
        self._sensor = sensor
        self._dim = dim
        self.sigma = sigma
        self.epsilon = epsilon

    def _compute_params(self, object_in_sensing_region):
        if object_in_sensing_region:
            # Object is in the sensing region
            alpha = self.epsilon
            beta = (1.0 - self.epsilon) / 2.0
            gamma = (1.0 - self.epsilon) / 2.0
        else:
            # Object is not in the sensing region.
            alpha = (1.0 - self.epsilon) / 2.0
            beta = (1.0 - self.epsilon) / 2.0
            gamma = self.epsilon
        return alpha, beta, gamma

    def probability(self, observation, next_state, action, **kwargs):
        """
        Returns the probability of Pr (observation | next_state, action).

        Args:
            observation (ObjectObservation)
            next_state (State)
            action (Action)
        """
        if not isinstance(action, LookAction):
            # No observation should be received
            if observation.pose == ObjectObservation.NULL:
                return 1.0
            else:
                return 0.0

        if observation.objid != self._objid:
            raise ValueError("The observation is not about the same object")

        # The (funny) business of allowing histogram belief update using O(oi|si',sr',a).
        next_robot_state = kwargs.get("next_robot_state", None)
        if next_robot_state is not None:
            assert next_robot_state["id"] == self._sensor.robot_id,\
                "Robot id of observation model mismatch with given state"
            robot_pose = next_robot_state.pose

            if isinstance(next_state, ObjectState):
                assert next_state["id"] == self._objid,\
                    "Object id of observation model mismatch with given state"
                object_pose = next_state.pose
            else:
                object_pose = next_state.pose(self._objid)
        else:
            robot_pose = next_state.pose(self._sensor.robot_id)
            object_pose = next_state.pose(self._objid)

        # Compute the probability
        zi = observation.pose
        alpha, beta, gamma = self._compute_params(self._sensor.within_range(robot_pose, object_pose))

        # Requires Python >= 3.6
        prob = 0.0
        # Event A:
        # object in sensing region and observation comes from object i
        if zi == ObjectObservation.NULL:
            # Even though event A occurred, the observation is NULL.
            # This has 0.0 probability.
            prob += 0.0 * alpha
        else:
            gaussian = pomdp_py.Gaussian(list(object_pose),
                                         [[self.sigma**2, 0],
                                          [0, self.sigma**2]])
            prob += gaussian[zi] * alpha

        # Event B
        prob += (1.0 / self._sensor.sensing_region_size) * beta

        # Event C
        pr_c = 1.0 if zi == ObjectObservation.NULL else 0.0  # indicator zi == NULL
        prob += pr_c * gamma
        return prob


    def sample(self, next_state, action, **kwargs):
        """Returns observation"""
        if not isinstance(action, LookAction):
            # Not a look action. So no observation
            return ObjectObservation(self._objid, ObjectObservation.NULL)

        robot_pose = next_state.pose(self._sensor.robot_id)
        object_pose = next_state.pose(self._objid)

        # Obtain observation according to distribution.
        alpha, beta, gamma = self._compute_params(self._sensor.within_range(robot_pose, object_pose))

        # Requires Python >= 3.6
        event_occured = random.choices(["A", "B", "C"], weights=[alpha, beta, gamma], k=1)[0]
        zi = self._sample_zi(event_occured, next_state)

        return ObjectObservation(self._objid, zi)

    def argmax(self, next_state, action, **kwargs):
        # Obtain observation according to distribution.
        alpha, beta, gamma = self._compute_params(self._sensor.within_range(robot_pose, object_pose))

        event_probs = {"A": alpha,
                       "B": beta,
                       "C": gamma}
        event_occured = max(event_probs, key=lambda e: event_probs[e])
        zi = self._sample_zi(event_occured, next_state, argmax=True)
        return ObjectObservation(self._objid, zi)

    def _sample_zi(self, event, next_state, argmax=False):
        if event == "A":
            object_true_pose = next_state.object_pose(self._objid)
            gaussian =  pomdp_py.Gaussian(list(object_true_pose),
                                          [[self.sigma**2, 0],
                                           [0, self.sigma**2]])
            if not argmax:
                zi = gaussian.random()
            else:
                zi = gaussian.mpe()
            zi = (int(round(zi[0])), int(round(zi[1])))

        elif event == "B":
            # TODO: FIX. zi should ONLY come from the field of view.
            # There is currently no easy way to sample from the field of view.
            width, height = self._dim
            zi = (random.randint(0, width),   # x axis
                  random.randint(0, height))  # y axis
        else: # event == C
            zi = ObjectObservation.NULL
        return zi


### Unit test ###
def unittest():
    from ..env.env import make_laser_sensor,\
        make_proximity_sensor, equip_sensors,\
        interpret, interpret_robot_id
    # Test within search region check,
    # and the observation model probability and
    # sampling functions.
    worldmap =\
        """
        ..........
        ....T.....
        ......x...
        ..T.r.T...
        ..x.......
        ....T.....
        ..........
        """
       #0123456789
       # 10 x 8
    worldstr = equip_sensors(worldmap,
                             {"r": make_laser_sensor(90, (1,5), 0.5, False)})
    env = interpret(worldstr)
    robot_id = interpret_robot_id("r")
    robot_pose = env.state.pose(robot_id)

    # within_range test
    sensor = env.sensors[robot_id]
    assert sensor.within_range(robot_pose, (4,3)) == False
    assert sensor.within_range(robot_pose, (5,3)) == True
    assert sensor.within_range(robot_pose, (6,3)) == True
    assert sensor.within_range(robot_pose, (7,2)) == True
    assert sensor.within_range(robot_pose, (7,3)) == True
    assert sensor.within_range(robot_pose, (4,3)) == False
    assert sensor.within_range(robot_pose, (2,4)) == False
    assert sensor.within_range(robot_pose, (4,1)) == False
    assert sensor.within_range(robot_pose, (4,5)) == False
    assert sensor.within_range(robot_pose, (0,0)) == False

    print(env.state)

    # observation model test
    O0 = ObjectObservationModel(0, sensor, (env.width, env.length), sigma=0.01, epsilon=1)
    O2 = ObjectObservationModel(2, sensor, (env.width, env.length), sigma=0.01, epsilon=1)
    O3 = ObjectObservationModel(3, sensor, (env.width, env.length), sigma=0.01, epsilon=1)
    O5 = ObjectObservationModel(5, sensor, (env.width, env.length), sigma=0.01, epsilon=1)

    z0 = O0.sample(env.state, Look)
    assert z0.pose == ObjectObservation.NULL
    z2 = O2.sample(env.state, Look)
    assert z2.pose == ObjectObservation.NULL
    z3 = O3.sample(env.state, Look)
    assert z3.pose == (6, 3)
    z5 = O5.sample(env.state, Look)
    assert z5.pose == ObjectObservation.NULL

    assert O0.probability(z0, env.state, Look) == 1.0
    assert O2.probability(z2, env.state, Look) == 1.0
    assert O3.probability(z3, env.state, Look) >= 1.0
    assert O3.probability(ObjectObservation(3, ObjectObservation.NULL),
                          env.state, Look) == 0.0
    assert O5.probability(z5, env.state, Look) == 1.0

if __name__ == "__main__":
    unittest()



from pomdp_py.framework.basics cimport GenerativeDistribution
import sys
import random as rand
import numpy as np

cdef class Histogram(GenerativeDistribution):

    """
    Histogram representation of a probability distribution.

    __init__(self, histogram)

    Args:
        histogram (dict) is a dictionary mapping from
            variable value to probability
    """

    def __init__(self, histogram):
        """`histogram` is a dictionary mapping from
        variable value to probability"""
        if not (isinstance(histogram, dict)):
            raise ValueError("Unsupported histogram representation! %s"
                             % str(type(histogram)))
        self._histogram = histogram

    @property
    def histogram(self):
        """histogram(self)"""
        return self._histogram

    def __str__(self):
        if isinstance(self._histogram, dict):
            return str([(k,self._histogram[k])
                        for k in list(reversed(sorted(self._histogram,
                                                      key=self._histogram.get)))[:5]])

    def __len__(self):
        return len(self._histogram)

    def __getitem__(self, value):
        """__getitem__(self, value)
        Returns the probability of `value`."""
        if value in self._histogram:
            return self._histogram[value]
        else:
            return 0

    def __setitem__(self, value, prob):
        """__setitem__(self, value, prob)
        Sets probability of value to `prob`."""
        self._histogram[value] = prob

    def __eq__(self, other):
        if not isinstance(other, Histogram):
            return False
        else:
            return self.histogram == other.histogram

    def __iter__(self):
        return iter(self._histogram)

    def mpe(self):
        """mpe(self)
        Returns the most likely value of the variable.
        """
        return max(self._histogram, key=self._histogram.get)

    def random(self, rnd=rand):
        """
        random(self)
        Randomly sample a value based on the probability
        in the histogram"""
        candidates = list(self._histogram.keys())
        prob_dist = []
        for value in candidates:
            prob_dist.append(self._histogram[value])
        if sys.version_info[0] >= 3 and sys.version_info[1] >= 6:
            # available in Python 3.6+
            return rnd.choices(candidates, weights=prob_dist, k=1)[0]
        else:
            print("Warning: Histogram.random() using np.random"\
                  "due to old Python version (<3.6). Random seed ignored.")
            return np.random.choice(candidates, 1, p=prob_dist)[0]

    def get_histogram(self):
        """get_histogram(self)
        Returns a dictionary from value to probability of the histogram"""
        return self._histogram

    # Deprecated; it's assuming non-log probabilities
    def is_normalized(self, epsilon=1e-9):
        """Returns true if this distribution is normalized"""
        prob_sum = sum(self._histogram[state] for state in self._histogram)
        return abs(1.0-prob_sum) < epsilon



"""The Environment"""

import pomdp_py
import copy
from pomdp_problems.multi_object_search.models.transition_model import *
from pomdp_problems.multi_object_search.models.reward_model import *
from pomdp_problems.multi_object_search.models.components.sensor import *
from pomdp_problems.multi_object_search.domain.state import *

class MosEnvironment(pomdp_py.Environment):
    """"""
    def __init__(self, dim, init_state, sensors, obstacles=set({})):
        """
        Args:
            sensors (dict): Map from robot_id to sensor (Sensor);
                            Sensors equipped on robots; Used to determine
                            which objects should be marked as found.
            obstacles (set): set of object ids that are obstacles;
                                The set difference of all object ids then
                                yields the target object ids."""
        self.width, self.length = dim
        self.sensors = sensors
        self.obstacles = obstacles
        transition_model = MosTransitionModel(dim,
                                              sensors,
                                              set(init_state.object_states.keys()))
        # Target objects, a set of ids, are not robot nor obstacles
        self.target_objects = \
            {objid
             for objid in set(init_state.object_states.keys()) - self.obstacles
             if not isinstance(init_state.object_states[objid], RobotState)}
        reward_model = GoalRewardModel(self.target_objects)
        super().__init__(init_state,
                         transition_model,
                         reward_model)
        
    @property
    def robot_ids(self):
        return set(self.sensors.keys())

    def state_transition(self, action, execute=True, robot_id=None):
        """state_transition(self, action, execute=True, **kwargs)

        Overriding parent class function.
        Simulates a state transition given `action`. If `execute` is set to True,
        then the resulting state will be the new current state of the environment.

        Args:
            action (Action): action that triggers the state transition
            execute (bool): If True, the resulting state of the transition will
                            become the current state.

        Returns:
            float or tuple: reward as a result of `action` and state
            transition, if `execute` is True (next_state, reward) if `execute`
            is False.

        """
        assert robot_id is not None, "state transition should happen for a specific robot"

        next_state = copy.deepcopy(self.state)
        next_state.object_states[robot_id] =\
            self.transition_model[robot_id].sample(self.state, action)
        
        reward = self.reward_model.sample(self.state, action, next_state,
                                          robot_id=robot_id)
        if execute:
            self.apply_transition(next_state)
            return reward
        else:
            return next_state, reward        

#### Interpret string as an initial world state ####
def interpret(worldstr):
    """
    Interprets a problem instance description in `worldstr`
    and returns the corresponding MosEnvironment.

    For example: This string
    
    .. code-block:: text

        rx...
        .x.xT
        .....
        ***
        r: laser fov=90 min_range=1 max_range=10
    
    describes a 3 by 5 world where x indicates obsticles and T indicates
    the "target object". T could be replaced by any upper-case letter A-Z
    which will serve as the object's id. Lower-case letters a-z (except for x)
    serve as id for robot(s).

    After the world, the :code:`***` signals description of the sensor for each robot.
    For example "r laser 90 1 10" means that robot `r` will have a Laser2Dsensor
    with fov 90, min_range 1.0, and max_range of 10.0.    

    Args:
        worldstr (str): a string that describes the initial state of the world.

    Returns:
        MosEnvironment: the corresponding environment for the world description.
            
    """
    worldlines = []
    sensorlines = []
    mode = "world"
    for line in worldstr.splitlines():
        line = line.strip()
        if len(line) > 0:
            if line == "***":
                mode = "sensor"
                continue
            if mode == "world":
                worldlines.append(line)
            if mode == "sensor":
                sensorlines.append(line)
    
    lines = [line for line in worldlines
             if len(line) > 0]
    w, l = len(worldlines[0]), len(worldlines)
    
    objects = {}    # objid -> object_state(pose)
    obstacles = set({})  # objid
    robots = {}  # robot_id -> robot_state(pose)
    sensors = {}  # robot_id -> Sensor

    # Parse world
    for y, line in enumerate(worldlines):
        if len(line) != w:
            raise ValueError("World size inconsistent."\
                             "Expected width: %d; Actual Width: %d"
                             % (w, len(line)))
        for x, c in enumerate(line):
            if c == "x":
                # obstacle
                objid = 1000 + len(obstacles)  # obstacle id
                objects[objid] = ObjectState(objid, "obstacle", (x,y))
                obstacles.add(objid)
                
            elif c.isupper():
                # target object
                objid = len(objects)
                objects[objid] = ObjectState(objid, "target", (x,y))
                
            elif c.islower():
                # robot
                robot_id = interpret_robot_id(c)
                robots[robot_id] = RobotState(robot_id, (x,y,0), (), None)

            else:
                assert c == ".", "Unrecognized character %s in worldstr" % c
    if len(robots) == 0:
        raise ValueError("No initial robot pose!")
    if len(objects) == 0:
        raise ValueError("No object!")

    # Parse sensors
    for line in sensorlines:
        if "," in line:
            raise ValueError("Wrong Fromat. SHould not have ','. Separate tokens with space.")
        robot_name = line.split(":")[0].strip()
        robot_id = interpret_robot_id(robot_name)
        assert robot_id in robots, "Sensor specified for unknown robot %s" % (robot_name)
        
        sensor_setting = line.split(":")[1].strip()
        sensor_type = sensor_setting.split(" ")[0].strip()
        sensor_params = {}
        for token in sensor_setting.split(" ")[1:]:
            param_name = token.split("=")[0].strip()
            param_value = eval(token.split("=")[1].strip())
            sensor_params[param_name] = param_value
        
        if sensor_type == "laser":
            sensor = Laser2DSensor(robot_id, **sensor_params)
        elif sensor_type == "proximity":
            sensor = ProximitySensor(robot_id, **sensor_params)
        else:
            raise ValueError("Unknown sensor type %s" % sensor_type)
        sensors[robot_id] = sensor

    return (w,l), robots, objects, obstacles, sensors

def interpret_robot_id(robot_name):
    return -ord(robot_name)


#### Utility functions for building the worldstr ####
def equip_sensors(worldmap, sensors):
    """
    Args:
        worldmap (str): a string that describes the initial state of the world.
        sensors (dict) a map from robot character representation (e.g. 'r') to a
    string that describes its sensor (e.g. 'laser fov=90 min_range=1 max_range=5
    angle_increment=5')

    Returns:
        str: A string that can be used as input to the `interpret` function
    """
    worldmap += "\n***\n"
    for robot_char in sensors:
        worldmap += "%s: %s\n" % (robot_char, sensors[robot_char])
    return worldmap

def make_laser_sensor(fov, dist_range, angle_increment, occlusion):
    """
    Returns string representation of the laser scanner configuration.
    For example:  "laser fov=90 min_range=1 max_range=10"

    Args:
        fov (int or float): angle between the start and end beams of one scan (degree).
        dist_range (tuple): (min_range, max_range)
        angle_increment (int or float): angular distance between measurements (rad).
        occlusion (bool): True if consider occlusion

    Returns:
        str: String representation of the laser scanner configuration.
    """
    fovstr = "fov=%s" % str(fov)
    rangestr = "min_range=%s max_range=%s" % (str(dist_range[0]), str(dist_range[1]))
    angicstr = "angle_increment=%s" % (str(angle_increment))
    occstr = "occlusion_enabled=%s" % str(occlusion)
    return "laser %s %s %s %s" % (fovstr, rangestr, angicstr, occstr)

def make_proximity_sensor(radius, occlusion):
    """
    Returns string representation of the proximity sensor configuration.
    For example: "proximity radius=5 occlusion_enabled=False"

    Args:
        radius (int or float)
        occlusion (bool): True if consider occlusion
    Returns:
        str: String representation of the proximity sensor configuration.
    """
    radiustr = "radius=%s" % str(radius)
    occstr = "occlusion_enabled=%s" % str(occlusion)
    return "proximity %s %s" % (radiustr, occstr)
